@@ -9,9 +9,12 @@
 """Helper functions for documentation, etc."""
 
 import inspect
+import io
 import logging
 import urllib.parse
 
+import wrapt
+
 logger = logging.getLogger(__name__)
 
 WORKAROUND_SCHEMES = ['s3', 's3n', 's3u', 's3a', 'gs']
@@ -74,7 +77,7 @@ def check_kwargs(kallable, kwargs):
     return supported_kwargs
 
 
-def clamp(value, minval, maxval):
+def clamp(value, minval=0, maxval=None):
     """Clamp a numeric value to a specific range.
 
     Parameters
@@ -94,7 +97,10 @@ def clamp(value, minval, maxval):
         The clamped value.  It will be in the range ``[minval, maxval]``.
 
     """
-    return max(min(value, maxval), minval)
+    if maxval is not None:
+        value = min(value, maxval)
+    value = max(value, minval)
+    return value
 
 
 def make_range_string(start=None, stop=None):
@@ -119,7 +125,9 @@ def make_range_string(start=None, stop=None):
     #
     if start is None and stop is None:
         raise ValueError("make_range_string requires either a stop or start value")
-    return 'bytes=%s-%s' % ('' if start is None else start, '' if stop is None else stop)
+    start_str = '' if start is None else str(start)
+    stop_str = '' if stop is None else str(stop)
+    return 'bytes=%s-%s' % (start_str, stop_str)
 
 
 def parse_content_range(content_range):
@@ -184,3 +192,29 @@ def safe_urlsplit(url):
 
     path = sr.path.replace(placeholder, '?')
     return urllib.parse.SplitResult(sr.scheme, sr.netloc, path, '', '')
+
+
+class TextIOWrapper(io.TextIOWrapper):
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        """Call close on underlying buffer only when there was no exception.
+
+        Without this patch, TextIOWrapper would call self.buffer.close() during
+        exception handling, which is unwanted for e.g. s3 and azure. They only call
+        self.close() when there was no exception (self.terminate() otherwise) to avoid
+        committing unfinished/failed uploads.
+        """
+        if exc_type is None:
+            self.close()
+
+
+class FileLikeProxy(wrapt.ObjectProxy):
+    def __init__(self, outer, inner):
+        super().__init__(outer)
+        self.__inner = inner
+
+    def __exit__(self, *args, **kwargs):
+        """Exit inner after exiting outer."""
+        try:
+            super().__exit__(*args, **kwargs)
+        finally:
+            self.__inner.__exit__(*args, **kwargs)
